REBOL [
	Title: "REBOL/Services - Async Message Queuing"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- Asynchronous Message Queuing ---------------------------------------------
;
;   These functions are used to queue message requests for transfer.
;   There are two queues in the session object:
;
;   send-queue: holds request objects that are waiting to be sent.
;   sent-queue; holds request objects that have been sent but not
;       yet replied.

req-msg-obj: context [
	port:       ; Port that it is queued on
	state:      ; States are [queued sent result aborted] or none.
	seq:        ; The request sequence num (used to id it)
	pri:        ; Priority of the message
	request:    ; The request block (payload)
	result:     ; The result (service reply) for the request
	callback:   ; Completion callback function (optional)
	wakeup:     ; When TRUE, causes WAIT function to return
		none
]

queue-request: func [
	"Create a request message, append to queue, and send if on top."
	port [port!]
	request [block!]
	callback [function! block! none!] "func [port result]"
	/local req queue
][
	queue: port/user-data/send-queue
	req: make req-msg-obj [
		state: 'queued
		seq: 0
		pri: 0
	]
	req/port: port
	req/request: request
	if block? :callback [callback: func [port result req] callback]
	req/callback: :callback

	; !!! Note:
	; This code might still present a race condition - the port
	; may be getting opened or closed at the same time.
	append queue req
	either all [web-port? port port/status = 'close] [
		; Re-open the port again:
		;reset-async-port port
		port/user-data/config/handler-context/opener port none
	][
		if req = first queue [send-request port]
	]
	req ; returned
]

send-request: func [
	"Encode and send the top queued request."
	port [port!]
	/local session req out
][
	; CRITICAL SECTION ------------------------------------------
	; Do not put print or IO of any kind in here.
	session: port/user-data
	req: session/send-queue/1
	; Must be an unsent request and the port must actually be open:
	if any [not req req/state = 'sent not open? port] [exit]
	req/state: 'sent
	;------------------------------------------------------------
	; Increment the sequence number:
	req/seq: session/seq: session/seq + 1
	debug request ['send-request req/seq]
	; Add standard request block header and sequence number:
	insert/only req/request reduce ['request 'seq req/seq]
	; Mold and send it:
	port/user-data/config/handler-context/sender port req/request
]

wake-request: func [
	"Deal with the server's reply from a request."
	port
	reply
	/local req seq
][
	if any [
		not block? reply
		reply/1 = 'fatal
	][ ; reply packet failed or worse
		debug error ['bad-reply mold reply]
		clear port/user-data/sent-queue ; !!! not good - abort them.
		do-error ["Service error:" mold reply]
		exit
	]
	req: port/user-data/sent-queue/1
	seq: reply-sequence? reply
	; Main callback always gets called (in case of unsolicited reply):
	port/user-data/callback port reply req
	; It better be the same request (same sequence #)
	either all [req req/seq = seq][
		remove port/user-data/sent-queue
		req/state: 'result
		req/result: reply
		req/callback port req/result req ; (callback may be NONE)
		;port/wakeup?: any [port/wakeup? found? req/wakeup]
		port/user-data/config/handler-context/set-wakeup port found? req/wakeup
	][
		debug error either req [
			['bad-sequence 'want req/seq 'got seq]
		][
			['no-queued-request]
		]
		clear port/user-data/sent-queue ; !!! not good - abort them.
		exit
	]
]

reply-sequence?: func [
	"Returns the packet's reply sequence number."
	reply
][
	all [
		block? reply
		block? reply/2
		select reply/2 'seq
	]
]

web-port?: func [port] [
	any [port/path port/target]
]
